/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jctools.queues.alt;

/**
 * @author nitsanw
 * 
 */
public interface ConcurrentQueueProducer<E> {
    /**
     * @param e
     * @return true if e was successfully offered, false if queue is full
     */
    boolean offer(E e);

    /**
     * @param e
     * @return true if e was successfully offered, false if not able to offer (queue may not be full)
     */
    boolean weakOffer(E e);
    
    int produce(ProducerFunction<E> p, int batchSize);
}